@@ -26,7 +26,7 @@ and will be converted automatically at the call site to the appropriate type.
 
 See [`@cfunction`](@ref).
 """
-struct CFunction <: Ref{Cvoid}
+mutable struct CFunction <: Ref{Cvoid}
     ptr::Ptr{Cvoid}
     f::Any
     _1::Ptr{Cvoid}
@@ -47,7 +47,7 @@ Note that the argument type tuple must be a literal tuple, and not a tuple-value
 (although it can include a splat expression). And that these arguments will be evaluated in global scope
 during compile-time (not deferred until runtime).
 Adding a '\\\$' in front of the function argument changes this to instead create a runtime closure
-over the local variable `callable`.
+over the local variable `callable` (this is not supported on all architectures).
 
 See [manual section on ccall and cfunction usage](@ref Calling-C-and-Fortran-Code).
 
@@ -61,12 +61,12 @@ julia> @cfunction(foo, Int, (Int, Int))
 Ptr{Cvoid} @0x000000001b82fcd0
 ```
 """
-macro cfunction(f, at, rt)
-    if !(isa(rt, Expr) && rt.head === :tuple)
+macro cfunction(f, rt, at)
+    if !(isa(at, Expr) && at.head === :tuple)
         throw(ArgumentError("@cfunction argument types must be a literal tuple"))
     end
-    rt.head = :call
-    pushfirst!(rt.args, GlobalRef(Core, :svec))
+    at.head = :call
+    pushfirst!(at.args, GlobalRef(Core, :svec))
     if isa(f, Expr) && f.head === :$
         fptr = f.args[1]
         typ = CFunction
@@ -74,7 +74,7 @@ macro cfunction(f, at, rt)
         fptr = QuoteNode(f)
         typ = Ptr{Cvoid}
     end
-    cfun = Expr(:cfunction, typ, fptr, at, rt, QuoteNode(:ccall))
+    cfun = Expr(:cfunction, typ, fptr, rt, at, QuoteNode(:ccall))
     return esc(cfun)
 end
 
@@ -161,8 +161,8 @@ end
 # construction from pointers
 Cstring(p::Union{Ptr{Int8},Ptr{UInt8},Ptr{Cvoid}}) = bitcast(Cstring, p)
 Cwstring(p::Union{Ptr{Cwchar_t},Ptr{Cvoid}})       = bitcast(Cwstring, p)
-(::Type{Ptr{T}})(p::Cstring) where {T<:Union{Int8,UInt8,Cvoid}} = bitcast(Ptr{T}, p)
-(::Type{Ptr{T}})(p::Cwstring) where {T<:Union{Cwchar_t,Cvoid}}  = bitcast(Ptr{Cwchar_t}, p)
+Ptr{T}(p::Cstring) where {T<:Union{Int8,UInt8,Cvoid}} = bitcast(Ptr{T}, p)
+Ptr{T}(p::Cwstring) where {T<:Union{Cwchar_t,Cvoid}}  = bitcast(Ptr{Cwchar_t}, p)
 
 convert(::Type{Cstring}, p::Union{Ptr{Int8},Ptr{UInt8},Ptr{Cvoid}}) = Cstring(p)
 convert(::Type{Cwstring}, p::Union{Ptr{Cwchar_t},Ptr{Cvoid}}) = Cwstring(p)
@@ -414,6 +414,18 @@ function transcode(::Type{UInt8}, src::AbstractVector{UInt16})
     return dst
 end
 
+function unsafe_string(p::Ptr{T}, length::Integer) where {T<:Union{UInt16,UInt32,Cwchar_t}}
+    transcode(String, unsafe_wrap(Array, p, length; own=false))
+end
+function unsafe_string(cw::Cwstring)
+    p = convert(Ptr{Cwchar_t}, cw)
+    n = 1
+    while unsafe_load(p, n) != 0
+        n += 1
+    end
+    return unsafe_string(p, n - 1)
+end
+
 # deferring (or un-deferring) ctrl-c handler for external C code that
 # is not interrupt safe (see also issue #2622).  The sigatomic_begin/end
 # functions should always be called in matched pairs, ideally via:
@@ -463,8 +475,27 @@ function reenable_sigint(f::Function)
     res
 end
 
-function ccallable(f::Function, rt::Type, argt::Type, name::Union{AbstractString,Symbol}=string(f))
-    ccall(:jl_extern_c, Cvoid, (Any, Any, Any, Cstring), f, rt, argt, name)
+"""
+    exit_on_sigint(on::Bool)
+
+Set `exit_on_sigint` flag of the julia runtime.  If `false`, Ctrl-C
+(SIGINT) is capturable as [`InterruptException`](@ref) in `try` block.
+This is the default behavior in REPL, any code run via `-e` and `-E`
+and in Julia script run with `-i` option.
+
+If `true`, `InterruptException` is not thrown by Ctrl-C.  Running code
+upon such event requires [`atexit`](@ref).  This is the default
+behavior in Julia script run without `-i` option.
+
+!!! compat "Julia 1.5"
+    Function `exit_on_sigint` requires at least Julia 1.5.
+"""
+function exit_on_sigint(on::Bool)
+    ccall(:jl_exit_on_sigint, Cvoid, (Cint,), on)
+end
+
+function _ccallable(rt::Type, sigt::Type)
+    ccall(:jl_extern_c, Cvoid, (Any, Any), rt, sigt)
 end
 
 function expand_ccallable(rt, def)
@@ -480,17 +511,22 @@ function expand_ccallable(rt, def)
             error("@ccallable requires a return type")
         end
         if sig.head === :call
-            name = sig.args[1]
+            f = sig.args[1]
+            if isa(f,Expr) && f.head === :(::)
+                f = f.args[end]
+            else
+                f = :(typeof($f))
+            end
             at = map(sig.args[2:end]) do a
                 if isa(a,Expr) && a.head === :(::)
-                    a.args[2]
+                    a.args[end]
                 else
                     :Any
                 end
             end
             return quote
                 $(esc(def))
-                ccallable($(esc(name)), $(esc(rt)), $(Expr(:curly, :Tuple, map(esc, at)...)), $(string(name)))
+                _ccallable($(esc(rt)), $(Expr(:curly, :Tuple, esc(f), map(esc, at)...)))
             end
         end
     end
@@ -503,3 +539,176 @@ end
 macro ccallable(rt, def)
     expand_ccallable(rt, def)
 end
+
+# @ccall implementation
+"""
+    ccall_macro_parse(expression)
+
+`ccall_macro_parse` is an implementation detail of `@ccall
+
+it takes an expression like `:(printf("%d"::Cstring, value::Cuint)::Cvoid)`
+returns: a tuple of `(function_name, return_type, arg_types, args)`
+
+The above input outputs this:
+
+    (:printf, :Cvoid, [:Cstring, :Cuint], ["%d", :value])
+"""
+function ccall_macro_parse(expr::Expr)
+    # setup and check for errors
+    if !Meta.isexpr(expr, :(::))
+        throw(ArgumentError("@ccall needs a function signature with a return type"))
+    end
+    rettype = expr.args[2]
+
+    call = expr.args[1]
+    if !Meta.isexpr(call, :call)
+        throw(ArgumentError("@ccall has to take a function call"))
+    end
+
+    # get the function symbols
+    func = let f = call.args[1]
+        if Meta.isexpr(f, :.)
+            :(($(f.args[2]), $(f.args[1])))
+        elseif Meta.isexpr(f, :$)
+            f
+        elseif f isa Symbol
+            QuoteNode(f)
+        else
+            throw(ArgumentError("@ccall function name must be a symbol, a `.` node (e.g. `libc.printf`) or an interpolated function pointer (with `\$`)"))
+        end
+    end
+
+    # detect varargs
+    varargs = nothing
+    argstart = 2
+    callargs = call.args
+    if length(callargs) >= 2 && Meta.isexpr(callargs[2], :parameters)
+        argstart = 3
+        varargs = callargs[2].args
+    end
+
+    # collect args and types
+    args = []
+    types = []
+
+    function pusharg!(arg)
+        if !Meta.isexpr(arg, :(::))
+            throw(ArgumentError("args in @ccall need type annotations. '$arg' doesn't have one."))
+        end
+        push!(args, arg.args[1])
+        push!(types, arg.args[2])
+    end
+
+    for i in argstart:length(callargs)
+        pusharg!(callargs[i])
+    end
+    # add any varargs if necessary
+    nreq = 0
+    if !isnothing(varargs)
+        if length(args) == 0
+            throw(ArgumentError("C ABI prohibits vararg without one required argument"))
+        end
+        nreq = length(args)
+        for a in varargs
+            pusharg!(a)
+        end
+    end
+
+    return func, rettype, types, args, nreq
+end
+
+
+function ccall_macro_lower(convention, func, rettype, types, args, nreq)
+    lowering = []
+    realargs = []
+    gcroots = []
+
+    # if interpolation was used, ensure  variable is a function pointer at runtime.
+    if Meta.isexpr(func, :$)
+        push!(lowering, Expr(:(=), :func, esc(func.args[1])))
+        name = QuoteNode(func.args[1])
+        func = :func
+        check = quote
+            if !isa(func, Ptr{Cvoid})
+                name = $name
+                throw(ArgumentError("interpolated function `$name` was not a Ptr{Cvoid}, but $(typeof(func))"))
+            end
+        end
+        push!(lowering, check)
+    else
+        func = esc(func)
+    end
+
+    for (i, (arg, type)) in enumerate(zip(args, types))
+        sym = Symbol(string("arg", i, "root"))
+        sym2 = Symbol(string("arg", i, ))
+        earg, etype = esc(arg), esc(type)
+        push!(lowering, :(local $sym = $(GlobalRef(Base, :cconvert))($etype, $earg)))
+        push!(lowering, :(local $sym2 = $(GlobalRef(Base, :unsafe_convert))($etype, $sym)))
+        push!(realargs, sym2)
+        push!(gcroots, sym)
+    end
+    etypes = Expr(:call, Expr(:core, :svec), types...)
+    exp = Expr(:foreigncall,
+               func,
+               esc(rettype),
+               esc(etypes),
+               nreq,
+               QuoteNode(convention),
+               realargs..., gcroots...)
+    push!(lowering, exp)
+
+    return Expr(:block, lowering...)
+end
+
+"""
+    @ccall library.function_name(argvalue1::argtype1, ...)::returntype
+    @ccall function_name(argvalue1::argtype1, ...)::returntype
+    @ccall \$function_pointer(argvalue1::argtype1, ...)::returntype
+
+Call a function in a C-exported shared library, specified by
+`library.function_name`, where `library` is a string constant or
+literal. The library may be omitted, in which case the `function_name`
+is resolved in the current process. Alternatively, `@ccall` may
+also be used to call a function pointer `\$function_pointer`, such as
+one returned by `dlsym`.
+
+Each `argvalue` to `@ccall` is converted to the corresponding
+`argtype`, by automatic insertion of calls to `unsafe_convert(argtype,
+cconvert(argtype, argvalue))`. (See also the documentation for
+[`unsafe_convert`](@ref Base.unsafe_convert) and [`cconvert`](@ref
+Base.cconvert) for further details.) In most cases, this simply
+results in a call to `convert(argtype, argvalue)`.
+
+
+# Examples
+
+    @ccall strlen(s::Cstring)::Csize_t
+
+This calls the C standard library function:
+
+    size_t strlen(char *)
+
+with a Julia variable named `s`. See also `ccall`.
+
+Varargs are supported with the following convention:
+
+    @ccall printf("%s = %d"::Cstring ; "foo"::Cstring, foo::Cint)::Cint
+
+The semicolon is used to separate required arguments (of which there
+must be at least one) from variadic arguments.
+
+Example using an external library:
+
+    # C signature of g_uri_escape_string:
+    # char *g_uri_escape_string(const char *unescaped, const char *reserved_chars_allowed, gboolean allow_utf8);
+
+    const glib = "libglib-2.0"
+    @ccall glib.g_uri_escape_string(my_uri::Cstring, ":/"::Cstring, true::Cint)::Cstring
+
+The string literal could also be used directly before the function
+name, if desired `"libglib-2.0".g_uri_escape_string(...`
+"""
+macro ccall(expr)
+    return ccall_macro_lower(:ccall, ccall_macro_parse(expr)...)
+end
