@@ -71,6 +71,17 @@ void JL_UV_LOCK(void)
     }
 }
 
+JL_DLLEXPORT void jl_iolock_begin(void)
+{
+    JL_UV_LOCK();
+}
+
+JL_DLLEXPORT void jl_iolock_end(void)
+{
+    JL_UV_UNLOCK();
+}
+
+
 void jl_uv_call_close_callback(jl_value_t *val)
 {
     jl_value_t *args[2];
@@ -94,11 +105,11 @@ static void jl_uv_closeHandle(uv_handle_t *handle)
         JL_STDERR = (JL_STREAM*)STDERR_FILENO;
     // also let the client app do its own cleanup
     if (handle->type != UV_FILE && handle->data) {
-        size_t last_age = jl_get_ptls_states()->world_age;
-        // TODO: data race on jl_world_counter across many files, to be fixed in a separate revision
-        jl_get_ptls_states()->world_age = jl_world_counter;
+        jl_ptls_t ptls = jl_get_ptls_states();
+        size_t last_age = ptls->world_age;
+        ptls->world_age = jl_world_counter;
         jl_uv_call_close_callback((jl_value_t*)handle->data);
-        jl_get_ptls_states()->world_age = last_age;
+        ptls->world_age = last_age;
     }
     if (handle == (uv_handle_t*)&signal_async)
         return;
@@ -146,7 +157,7 @@ static void uv_flush_callback(uv_write_t *req, int status)
 // Turn a normal write into a blocking write (primarily for use from C and gdb).
 // Warning: This calls uv_run, so it can have unbounded side-effects.
 // Be care where you call it from! - the libuv loop is also not reentrant.
-void jl_uv_flush(uv_stream_t *stream)
+JL_DLLEXPORT void jl_uv_flush(uv_stream_t *stream)
 {
     if (stream == (void*)STDIN_FILENO ||
         stream == (void*)STDOUT_FILENO ||
@@ -162,7 +173,7 @@ void jl_uv_flush(uv_stream_t *stream)
         uv_buf_t buf;
         buf.base = (char*)(&buf + 1);
         buf.len = 0;
-        uv_write_t *write_req = (uv_write_t*)malloc(sizeof(uv_write_t));
+        uv_write_t *write_req = (uv_write_t*)malloc_s(sizeof(uv_write_t));
         write_req->data = (void*)&fired;
         if (uv_write(write_req, stream, &buf, 1, uv_flush_callback) != 0) {
             JL_UV_UNLOCK();
@@ -192,28 +203,15 @@ JL_DLLEXPORT void *jl_uv_write_handle(uv_write_t *req) { return req->handle; }
 
 extern volatile unsigned _threadedregion;
 
-JL_DLLEXPORT int jl_run_once(uv_loop_t *loop)
-{
-    jl_ptls_t ptls = jl_get_ptls_states();
-    if (loop) {
-        jl_gc_safepoint_(ptls);
-        JL_UV_LOCK();
-        loop->stop_flag = 0;
-        int r = uv_run(loop,UV_RUN_ONCE);
-        JL_UV_UNLOCK();
-        return r;
-    }
-    return 0;
-}
-
-JL_DLLEXPORT int jl_process_events(uv_loop_t *loop)
+JL_DLLEXPORT int jl_process_events(void)
 {
     jl_ptls_t ptls = jl_get_ptls_states();
+    uv_loop_t *loop = jl_io_loop;
     if (loop && (_threadedregion || ptls->tid == 0)) {
         jl_gc_safepoint_(ptls);
-        if (jl_mutex_trylock(&jl_uv_mutex)) {
+        if (jl_atomic_load(&jl_uv_n_waiters) == 0 && jl_mutex_trylock(&jl_uv_mutex)) {
             loop->stop_flag = 0;
-            int r = uv_run(loop,UV_RUN_NOWAIT);
+            int r = uv_run(loop, UV_RUN_NOWAIT);
             JL_UV_UNLOCK();
             return r;
         }
@@ -249,7 +247,7 @@ JL_DLLEXPORT void jl_close_uv(uv_handle_t *handle)
     }
 
     if (handle->type == UV_NAMED_PIPE || handle->type == UV_TCP || handle->type == UV_TTY) {
-        uv_write_t *req = (uv_write_t*)malloc(sizeof(uv_write_t));
+        uv_write_t *req = (uv_write_t*)malloc_s(sizeof(uv_write_t));
         req->handle = (uv_stream_t*)handle;
         jl_uv_flush_close_callback(req, 0);
         JL_UV_UNLOCK();
@@ -268,7 +266,9 @@ JL_DLLEXPORT void jl_forceclose_uv(uv_handle_t *handle)
     // avoid double-closing the stream
     if (!uv_is_closing(handle)) {
         JL_UV_LOCK();
-        uv_close(handle, &jl_uv_closeHandle);
+        if (!uv_is_closing(handle)) {
+            uv_close(handle, &jl_uv_closeHandle);
+        }
         JL_UV_UNLOCK();
     }
 }
@@ -392,8 +392,16 @@ JL_DLLEXPORT int jl_fs_chown(char *path, int uid, int gid)
     return ret;
 }
 
+JL_DLLEXPORT int jl_fs_access(char *path, int mode)
+{
+    uv_fs_t req;
+    int ret = uv_fs_access(unused_uv_loop_arg, &req, path, mode, NULL);
+    uv_fs_req_cleanup(&req);
+    return ret;
+}
+
 JL_DLLEXPORT int jl_fs_write(uv_os_fd_t handle, const char *data, size_t len,
-                             int64_t offset)
+                             int64_t offset) JL_NOTSAFEPOINT
 {
     jl_ptls_t ptls = jl_get_ptls_states();
     // TODO: fix this cheating
@@ -519,9 +527,9 @@ JL_DLLEXPORT void jl_uv_puts(uv_stream_t *stream, const char *str, size_t n)
     }
     else {
         // Write to libuv stream...
-        uv_write_t *req = (uv_write_t*)malloc(sizeof(uv_write_t)+n);
-        char *data = (char*)(req+1);
-        memcpy(data,str,n);
+        uv_write_t *req = (uv_write_t*)malloc_s(sizeof(uv_write_t) + n);
+        char *data = (char*)(req + 1);
+        memcpy(data, str, n);
         uv_buf_t buf[1];
         buf[0].base = data;
         buf[0].len = n;
@@ -563,10 +571,10 @@ extern int vasprintf(char **str, const char *fmt, va_list ap);
 
 JL_DLLEXPORT int jl_vprintf(uv_stream_t *s, const char *format, va_list args)
 {
-    char *str=NULL;
+    char *str = NULL;
     int c;
     va_list al;
-#if defined(_OS_WINDOWS_) && !defined(_COMPILER_MINGW_)
+#if defined(_OS_WINDOWS_) && !defined(_COMPILER_GCC_)
     al = args;
 #else
     va_copy(al, args);
@@ -593,10 +601,14 @@ JL_DLLEXPORT int jl_printf(uv_stream_t *s, const char *format, ...)
     return c;
 }
 
-JL_DLLEXPORT void jl_safe_printf(const char *fmt, ...)
+JL_DLLEXPORT void jl_safe_printf(const char *fmt, ...) JL_NOTSAFEPOINT
 {
     static char buf[1000];
     buf[0] = '\0';
+    int last_errno = errno;
+#ifdef _OS_WINDOWS_
+    DWORD last_error = GetLastError();
+#endif
 
     va_list args;
     va_start(args, fmt);
@@ -608,6 +620,10 @@ JL_DLLEXPORT void jl_safe_printf(const char *fmt, ...)
     if (write(STDERR_FILENO, buf, strlen(buf)) < 0) {
         // nothing we can do; ignore the failure
     }
+#ifdef _OS_WINDOWS_
+    SetLastError(last_error);
+#endif
+    errno = last_errno;
 }
 
 JL_DLLEXPORT void jl_exit(int exitcode)
@@ -617,7 +633,7 @@ JL_DLLEXPORT void jl_exit(int exitcode)
     exit(exitcode);
 }
 
-JL_DLLEXPORT int jl_getpid(void)
+JL_DLLEXPORT int jl_getpid(void) JL_NOTSAFEPOINT
 {
 #ifdef _OS_WINDOWS_
     return GetCurrentProcessId();
@@ -626,28 +642,33 @@ JL_DLLEXPORT int jl_getpid(void)
 #endif
 }
 
-//NOTE: These function expects port/host to be in network byte-order (Big Endian)
-JL_DLLEXPORT int jl_tcp_bind(uv_tcp_t *handle, uint16_t port, uint32_t host,
-                             unsigned int flags)
+typedef union {
+    struct sockaddr in;
+    struct sockaddr_in v4;
+    struct sockaddr_in6 v6;
+} uv_sockaddr_in;
+
+static void jl_sockaddr_fill(uv_sockaddr_in *addr, uint16_t port, void *host, int ipv6)
 {
-    struct sockaddr_in addr;
-    memset(&addr, 0, sizeof(struct sockaddr_in));
-    addr.sin_port = port;
-    addr.sin_addr.s_addr = host;
-    addr.sin_family = AF_INET;
-    // TODO: do we need a lock here?
-    return uv_tcp_bind(handle, (struct sockaddr*)&addr, flags);
+    memset(addr, 0, sizeof(*addr));
+    if (ipv6) {
+        addr->v6.sin6_family = AF_INET6;
+        memcpy(&addr->v6.sin6_addr, host, 16);
+        addr->v6.sin6_port = port;
+    }
+    else {
+        addr->v4.sin_family = AF_INET;
+        addr->v4.sin_addr.s_addr = *(uint32_t*)host;
+        addr->v4.sin_port = port;
+    }
 }
 
-JL_DLLEXPORT int jl_tcp_bind6(uv_tcp_t *handle, uint16_t port, void *host,
-                              unsigned int flags)
+//NOTE: These function expects port/host to be in network byte-order (Big Endian)
+JL_DLLEXPORT int jl_tcp_bind(uv_tcp_t *handle, uint16_t port, void *host,
+                             unsigned int flags, int ipv6)
 {
-    struct sockaddr_in6 addr;
-    memset(&addr, 0, sizeof(struct sockaddr_in6));
-    addr.sin6_port = port;
-    memcpy(&addr.sin6_addr, host, 16);
-    addr.sin6_family = AF_INET6;
-    // TODO: do we need a lock here
+    uv_sockaddr_in addr;
+    jl_sockaddr_fill(&addr, port, host, ipv6);
     return uv_tcp_bind(handle, (struct sockaddr*)&addr, flags);
 }
 
@@ -659,6 +680,8 @@ JL_DLLEXPORT int jl_tcp_getsockname(uv_tcp_t *handle, uint16_t *port,
     memset(&addr, 0, sizeof(struct sockaddr_storage));
     namelen = sizeof addr;
     int res = uv_tcp_getsockname(handle, (struct sockaddr*)&addr, &namelen);
+    if (res)
+        return res;
     *family = addr.ss_family;
     if (addr.ss_family == AF_INET) {
         struct sockaddr_in *addr4 = (struct sockaddr_in*)&addr;
@@ -670,9 +693,6 @@ JL_DLLEXPORT int jl_tcp_getsockname(uv_tcp_t *handle, uint16_t *port,
         *port = addr6->sin6_port;
         memcpy(host, &(addr6->sin6_addr), 16);
     }
-    else {
-        return -1;
-    }
     return res;
 }
 
@@ -684,6 +704,8 @@ JL_DLLEXPORT int jl_tcp_getpeername(uv_tcp_t *handle, uint16_t *port,
     memset(&addr, 0, sizeof(struct sockaddr_storage));
     namelen = sizeof addr;
     int res = uv_tcp_getpeername(handle, (struct sockaddr*)&addr, &namelen);
+    if (res)
+        return res;
     *family = addr.ss_family;
     if (addr.ss_family == AF_INET) {
         struct sockaddr_in *addr4 = (struct sockaddr_in*)&addr;
@@ -695,69 +717,26 @@ JL_DLLEXPORT int jl_tcp_getpeername(uv_tcp_t *handle, uint16_t *port,
         *port = addr6->sin6_port;
         memcpy(host, &(addr6->sin6_addr), 16);
     }
-    else {
-        return -1;
-    }
     return res;
 }
 
-JL_DLLEXPORT int jl_udp_bind(uv_udp_t *handle, uint16_t port, uint32_t host,
-                             uint32_t flags)
+JL_DLLEXPORT int jl_udp_bind(uv_udp_t *handle, uint16_t port, void *host,
+                             uint32_t flags, int ipv6)
 {
-    struct sockaddr_in addr;
-    memset(&addr, 0, sizeof(struct sockaddr_in));
-    addr.sin_port = port;
-    addr.sin_addr.s_addr = host;
-    addr.sin_family = AF_INET;
+    uv_sockaddr_in addr;
+    jl_sockaddr_fill(&addr, port, host, ipv6);
     return uv_udp_bind(handle, (struct sockaddr*)&addr, flags);
 }
 
-JL_DLLEXPORT int jl_udp_bind6(uv_udp_t *handle, uint16_t port, void *host,
-                              uint32_t flags)
+JL_DLLEXPORT int jl_udp_send(uv_udp_send_t *req, uv_udp_t *handle, uint16_t port, void *host,
+                             char *data, uint32_t size, uv_udp_send_cb cb, int ipv6)
 {
-    struct sockaddr_in6 addr;
-    memset(&addr, 0, sizeof(struct sockaddr_in6));
-    addr.sin6_port = port;
-    memcpy(&addr.sin6_addr, host, 16);
-    addr.sin6_family = AF_INET6;
-    return uv_udp_bind(handle, (struct sockaddr*)&addr, flags);
-}
-
-JL_DLLEXPORT int jl_udp_send(uv_udp_t *handle, uint16_t port, uint32_t host,
-                             void *data, uint32_t size, uv_udp_send_cb cb)
-{
-    struct sockaddr_in addr;
-    memset(&addr, 0, sizeof(struct sockaddr_in));
-    addr.sin_port = port;
-    addr.sin_addr.s_addr = host;
-    addr.sin_family = AF_INET;
+    uv_sockaddr_in addr;
+    jl_sockaddr_fill(&addr, port, host, ipv6);
     uv_buf_t buf[1];
-    buf[0].base = (char *) data;
+    buf[0].base = data;
     buf[0].len = size;
-    uv_udp_send_t *req = (uv_udp_send_t*)malloc(sizeof(uv_udp_send_t));
-    req->data = handle->data;
-    JL_UV_LOCK();
     int r = uv_udp_send(req, handle, buf, 1, (struct sockaddr*)&addr, cb);
-    JL_UV_UNLOCK();
-    return r;
-}
-
-JL_DLLEXPORT int jl_udp_send6(uv_udp_t *handle, uint16_t port, void *host,
-                              void *data, uint32_t size, uv_udp_send_cb cb)
-{
-    struct sockaddr_in6 addr;
-    memset(&addr, 0, sizeof(struct sockaddr_in6));
-    addr.sin6_port = port;
-    memcpy(&addr.sin6_addr, host, 16);
-    addr.sin6_family = AF_INET6;
-    uv_buf_t buf[1];
-    buf[0].base = (char *) data;
-    buf[0].len = size;
-    uv_udp_send_t *req = (uv_udp_send_t *) malloc(sizeof(uv_udp_send_t));
-    req->data = handle->data;
-    JL_UV_LOCK();
-    int r = uv_udp_send(req, handle, buf, 1, (struct sockaddr*)&addr, cb);
-    JL_UV_UNLOCK();
     return r;
 }
 
@@ -769,7 +748,7 @@ JL_DLLEXPORT int jl_uv_sizeof_interface_address(void)
 JL_DLLEXPORT int jl_uv_interface_addresses(uv_interface_address_t **ifAddrStruct,
                                            int *count)
 {
-    return uv_interface_addresses(ifAddrStruct,count);
+    return uv_interface_addresses(ifAddrStruct, count);
 }
 
 JL_DLLEXPORT int jl_uv_interface_address_is_internal(uv_interface_address_t *addr)
@@ -796,124 +775,73 @@ JL_DLLEXPORT int jl_getaddrinfo(uv_loop_t *loop, uv_getaddrinfo_t *req,
 }
 
 JL_DLLEXPORT int jl_getnameinfo(uv_loop_t *loop, uv_getnameinfo_t *req,
-        uint32_t host, uint16_t port, int flags, uv_getnameinfo_cb uvcb)
+        void *host, uint16_t port, int flags, uv_getnameinfo_cb uvcb, int ipv6)
 {
-    struct sockaddr_in addr;
-    memset(&addr, 0, sizeof(addr));
-    addr.sin_family = AF_INET;
-    addr.sin_addr.s_addr = host;
-    addr.sin_port = port;
-
-    req->data = NULL;
+    uv_sockaddr_in addr;
+    jl_sockaddr_fill(&addr, port, host, ipv6);
     return uv_getnameinfo(loop, req, uvcb, (struct sockaddr*)&addr, flags);
 }
 
-JL_DLLEXPORT int jl_getnameinfo6(uv_loop_t *loop, uv_getnameinfo_t *req,
-        void *host, uint16_t port, int flags, uv_getnameinfo_cb uvcb)
-{
-    struct sockaddr_in6 addr;
-    memset(&addr, 0, sizeof(addr));
-    addr.sin6_family = AF_INET6;
-    memcpy(&addr.sin6_addr, host, 16);
-    addr.sin6_port = port;
-
-    req->data = NULL;
-    JL_UV_LOCK();
-    int r = uv_getnameinfo(loop, req, uvcb, (struct sockaddr*)&addr, flags);
-    JL_UV_UNLOCK();
-    return r;
-}
-
-
 JL_DLLEXPORT struct sockaddr *jl_sockaddr_from_addrinfo(struct addrinfo *addrinfo)
 {
     return addrinfo->ai_addr;
 }
+
 JL_DLLEXPORT struct addrinfo *jl_next_from_addrinfo(struct addrinfo *addrinfo)
 {
     return addrinfo->ai_next;
 }
 
-JL_DLLEXPORT int jl_sockaddr_in_is_ip4(struct sockaddr_in *addr)
+JL_DLLEXPORT int jl_sockaddr_is_ip4(struct sockaddr *addr)
 {
-    return (addr->sin_family==AF_INET);
+    return (addr->sa_family == AF_INET);
 }
 
-JL_DLLEXPORT int jl_sockaddr_in_is_ip6(struct sockaddr_in *addr)
+JL_DLLEXPORT int jl_sockaddr_is_ip6(struct sockaddr *addr)
 {
-    return (addr->sin_family==AF_INET6);
+    return (addr->sa_family == AF_INET6);
 }
 
-JL_DLLEXPORT int jl_sockaddr_is_ip4(struct sockaddr_storage *addr)
-{
-    return (addr->ss_family==AF_INET);
-}
-
-JL_DLLEXPORT int jl_sockaddr_is_ip6(struct sockaddr_storage *addr)
-{
-    return (addr->ss_family==AF_INET6);
-}
-
-JL_DLLEXPORT unsigned int jl_sockaddr_host4(struct sockaddr_in *addr)
+JL_DLLEXPORT uint32_t jl_sockaddr_host4(struct sockaddr_in *addr)
 {
     return addr->sin_addr.s_addr;
 }
 
-JL_DLLEXPORT unsigned int jl_sockaddr_host6(struct sockaddr_in6 *addr, char *host)
+JL_DLLEXPORT unsigned jl_sockaddr_host6(struct sockaddr_in6 *addr, char *host)
 {
     memcpy(host, &addr->sin6_addr, 16);
     return addr->sin6_scope_id;
 }
 
-JL_DLLEXPORT void jl_sockaddr_set_port(struct sockaddr_storage *addr,
-                                       uint16_t port)
+JL_DLLEXPORT uint16_t jl_sockaddr_port4(struct sockaddr_in *addr)
 {
-    if (addr->ss_family==AF_INET)
-        ((struct sockaddr_in*)addr)->sin_port = port;
+    return addr->sin_port;
+}
+
+JL_DLLEXPORT uint16_t jl_sockaddr_port6(struct sockaddr_in6 *addr)
+{
+    return addr->sin6_port;
+}
+
+
+JL_DLLEXPORT void jl_sockaddr_set_port(uv_sockaddr_in *addr, uint16_t port)
+{
+    if (addr->in.sa_family == AF_INET)
+        addr->v4.sin_port = port;
     else
-        ((struct sockaddr_in6*)addr)->sin6_port = port;
+        addr->v6.sin6_port = port;
 }
 
-JL_DLLEXPORT int jl_tcp4_connect(uv_tcp_t *handle,uint32_t host, uint16_t port,
-                                 uv_connect_cb cb)
+JL_DLLEXPORT int jl_tcp_connect(uv_tcp_t *handle, void *host, uint16_t port,
+                                uv_connect_cb cb, int ipv6)
 {
-    struct sockaddr_in addr;
-    uv_connect_t *req = (uv_connect_t*)malloc(sizeof(uv_connect_t));
-    req->data = 0;
-    memset(&addr, 0, sizeof(struct sockaddr_in));
-    addr.sin_family = AF_INET;
-    addr.sin_addr.s_addr = host;
-    addr.sin_port = port;
-    JL_UV_LOCK();
-    int r = uv_tcp_connect(req,handle,(struct sockaddr*)&addr,cb);
-    JL_UV_UNLOCK();
-    return r;
-}
-
-JL_DLLEXPORT int jl_tcp6_connect(uv_tcp_t *handle, void *host, uint16_t port,
-                                 uv_connect_cb cb)
-{
-    struct sockaddr_in6 addr;
-    uv_connect_t *req = (uv_connect_t*)malloc(sizeof(uv_connect_t));
-    req->data = 0;
-    memset(&addr, 0, sizeof(struct sockaddr_in6));
-    addr.sin6_family = AF_INET6;
-    memcpy(&addr.sin6_addr, host, 16);
-    addr.sin6_port = port;
-    JL_UV_LOCK();
-    int r = uv_tcp_connect(req,handle,(struct sockaddr*)&addr,cb);
-    JL_UV_UNLOCK();
-    return r;
-}
-
-JL_DLLEXPORT int jl_connect_raw(uv_tcp_t *handle,struct sockaddr_storage *addr,
-                                uv_connect_cb cb)
-{
-    uv_connect_t *req = (uv_connect_t*)malloc(sizeof(uv_connect_t));
-    req->data = 0;
-    JL_UV_LOCK();
-    int r = uv_tcp_connect(req,handle,(struct sockaddr*)addr,cb);
-    JL_UV_UNLOCK();
+    uv_sockaddr_in addr;
+    jl_sockaddr_fill(&addr, port, host, ipv6);
+    uv_connect_t *req = (uv_connect_t*)malloc_s(sizeof(uv_connect_t));
+    req->data = NULL;
+    int r = uv_tcp_connect(req, handle, &addr.in, cb);
+    if (r)
+        free(req);
     return r;
 }
 
@@ -1046,7 +974,6 @@ struct work_baton {
     void      *work_args;
     void      *work_retval;
     notify_cb_t notify_func;
-    int       tid;
     int       notify_idx;
 };
 
@@ -1070,7 +997,7 @@ void jl_work_notifier(uv_work_t *req, int status)
 JL_DLLEXPORT int jl_queue_work(work_cb_t work_func, void *work_args, void *work_retval,
                                notify_cb_t notify_func, int notify_idx)
 {
-    struct work_baton *baton = (struct work_baton*) malloc(sizeof(struct work_baton));
+    struct work_baton *baton = (struct work_baton*)malloc_s(sizeof(struct work_baton));
     baton->req.data = (void*) baton;
     baton->work_func = work_func;
     baton->work_args = work_args;
@@ -1085,101 +1012,6 @@ JL_DLLEXPORT int jl_queue_work(work_cb_t work_func, void *work_args, void *work_
     return 0;
 }
 
-JL_DLLEXPORT void jl_uv_stop(uv_loop_t* loop)
-{
-    JL_UV_LOCK();
-    uv_stop(loop);
-    // TODO: use memory/compiler fence here instead of the lock
-    JL_UV_UNLOCK();
-}
-
-JL_DLLEXPORT void jl_uv_update_time(uv_loop_t* loop)
-{
-    JL_UV_LOCK();
-    uv_update_time(loop);
-    JL_UV_UNLOCK();
-}
-
-JL_DLLEXPORT int jl_uv_timer_start(uv_timer_t* handle, uv_timer_cb cb,
-                                   uint64_t timeout, uint64_t repeat)
-{
-    JL_UV_LOCK();
-    int r = uv_timer_start(handle, cb, timeout, repeat);
-    JL_UV_UNLOCK();
-    return r;
-}
-
-JL_DLLEXPORT int jl_uv_timer_stop(uv_timer_t* handle)
-{
-    JL_UV_LOCK();
-    int r = uv_timer_stop(handle);
-    JL_UV_UNLOCK();
-    return r;
-}
-
-JL_DLLEXPORT int jl_uv_fs_scandir(uv_loop_t* loop, uv_fs_t* req, const char* path, int flags,
-                                  uv_fs_cb cb)
-{
-    JL_UV_LOCK();
-    int r = uv_fs_scandir(loop, req, path, flags, cb);
-    JL_UV_UNLOCK();
-    return r;
-}
-
-JL_DLLEXPORT int jl_uv_fs_readlink(uv_loop_t* loop, uv_fs_t* req, const char* path,
-                                   uv_fs_cb cb)
-{
-    JL_UV_LOCK();
-    int r = uv_fs_readlink(loop, req, path, cb);
-    JL_UV_UNLOCK();
-    return r;
-}
-
-JL_DLLEXPORT int jl_uv_fs_open(uv_loop_t* loop, uv_fs_t* req, const char* path, int flags,
-                               int mode, uv_fs_cb cb)
-{
-    JL_UV_LOCK();
-    int r = uv_fs_open(loop, req, path, flags, mode, cb);
-    JL_UV_UNLOCK();
-    return r;
-}
-
-JL_DLLEXPORT int jl_uv_fs_ftruncate(uv_loop_t* loop, uv_fs_t* req, uv_os_fd_t handle,
-                                    int64_t offset, uv_fs_cb cb)
-{
-    JL_UV_LOCK();
-    int r = uv_fs_ftruncate(loop, req, handle, offset, cb);
-    JL_UV_UNLOCK();
-    return r;
-}
-
-JL_DLLEXPORT int jl_uv_fs_futime(uv_loop_t* loop, uv_fs_t* req, uv_os_fd_t handle, double atime,
-                                 double mtime, uv_fs_cb cb)
-{
-    JL_UV_LOCK();
-    int r = uv_fs_futime(loop, req, handle, atime, mtime, cb);
-    JL_UV_UNLOCK();
-    return r;
-}
-
-JL_DLLEXPORT int jl_uv_read_start(uv_stream_t* handle, uv_alloc_cb alloc_cb,
-                                  uv_read_cb read_cb)
-{
-    JL_UV_LOCK();
-    int r = uv_read_start(handle, alloc_cb, read_cb);
-    JL_UV_UNLOCK();
-    return r;
-}
-
-JL_DLLEXPORT int jl_uv_read_stop(uv_stream_t* handle)
-{
-    JL_UV_LOCK();
-    int r = uv_read_stop(handle);
-    JL_UV_UNLOCK();
-    return r;
-}
-
-
 #ifndef _OS_WINDOWS_
 #if defined(__APPLE__)
 int uv___stream_fd(uv_stream_t *handle);
